# Metadata for loading pretrained model weights.
backbone_presets = {
    "moonshine_tiny_en": {
        "metadata": {
            "description": (
                "Moonshine tiny model for English speech recognition. "
                "Developed by Useful Sensors for real-time transcription."
            ),
            "params": 27092736,
            "path": "moonshine",
        },
        "kaggle_handle": "kaggle://keras/moonshine/Keras/moonshine_tiny_en/1",
    },
    "moonshine_base_en": {
        "metadata": {
            "description": (
                "Moonshine base model for English speech recognition. "
                "Developed by Useful Sensors for real-time transcription."
            ),
            "params": 61513920,
            "path": "moonshine",
        },
        "kaggle_handle": "kaggle://keras/moonshine/Keras/moonshine_base_en/1",
    },
}
