# Metadata for loading pretrained model weights.
backbone_presets = {
    "t5gemma_s_s_ul2": {
        "metadata": {
            "description": (
                "T5Gemma S/S model with a small encoder and small decoder, "
                "adapted as a UL2 model."
            ),
            "params": 312517632,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_s_s_ul2/1",
    },
    "t5gemma_s_s_prefixlm": {
        "metadata": {
            "description": (
                "T5Gemma S/S model with a small encoder and small decoder, "
                "adapted as a prefix language model."
            ),
            "params": 312517632,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_s_s_prefixlm/1",
    },
    "t5gemma_s_s_ul2_it": {
        "metadata": {
            "description": (
                "T5Gemma S/S model with a small encoder and small decoder, "
                "adapted as a UL2 model and fine-tuned for instruction "
                "following."
            ),
            "params": 312517632,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_s_s_ul2_it/1",
    },
    "t5gemma_s_s_prefixlm_it": {
        "metadata": {
            "description": (
                "T5Gemma S/S model with a small encoder and small decoder, "
                "adapted as a prefix language model and fine-tuned for "
                "instruction following."
            ),
            "params": 312517632,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_s_s_prefixlm_it/1",
    },
    "t5gemma_b_b_ul2": {
        "metadata": {
            "description": (
                "T5Gemma B/B model with a base encoder and base decoder, "
                "adapted as a UL2 model."
            ),
            "params": 591490560,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_b_b_ul2/1",
    },
    "t5gemma_b_b_prefixlm": {
        "metadata": {
            "description": (
                "T5Gemma B/B model with a base encoder and base decoder, "
                "adapted as a prefix language model."
            ),
            "params": 591490560,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_b_b_prefixlm/1",
    },
    "t5gemma_b_b_ul2_it": {
        "metadata": {
            "description": (
                "T5Gemma B/B model with a base encoder and base decoder, "
                "adapted as a UL2 model and fine-tuned for instruction "
                "following."
            ),
            "params": 591490560,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_b_b_ul2_it/1",
    },
    "t5gemma_b_b_prefixlm_it": {
        "metadata": {
            "description": (
                "T5Gemma B/B model with a base encoder and base decoder, "
                "adapted as a prefix language model and fine-tuned for "
                "instruction following."
            ),
            "params": 591490560,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_b_b_prefixlm_it/1",
    },
    "t5gemma_l_l_ul2": {
        "metadata": {
            "description": (
                "T5Gemma L/L model with a large encoder and large decoder, "
                "adapted as a UL2 model."
            ),
            "params": 1241761792,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_l_l_ul2/1",
    },
    "t5gemma_l_l_prefixlm": {
        "metadata": {
            "description": (
                "T5Gemma L/L model with a large encoder and large decoder, "
                "adapted as a prefix language model."
            ),
            "params": 1241761792,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_l_l_prefixlm/1",
    },
    "t5gemma_l_l_ul2_it": {
        "metadata": {
            "description": (
                "T5Gemma L/L model with a large encoder and large decoder, "
                "adapted as a UL2 model and fine-tuned for instruction "
                "following."
            ),
            "params": 1241761792,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_l_l_ul2_it/1",
    },
    "t5gemma_l_l_prefixlm_it": {
        "metadata": {
            "description": (
                "T5Gemma L/L model with a large encoder and large decoder, "
                "adapted as a prefix language model and fine-tuned for "
                "instruction following."
            ),
            "params": 1241761792,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_l_l_prefixlm_it/1",
    },
    "t5gemma_ml_ml_ul2": {
        "metadata": {
            "description": (
                "T5Gemma ML/ML model with a medium-large encoder and "
                "medium-large decoder, adapted as a UL2 model."
            ),
            "params": 2200345344,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_ml_ml_ul2/1",
    },
    "t5gemma_ml_ml_prefixlm": {
        "metadata": {
            "description": (
                "T5Gemma ML/ML model with a medium-large encoder and "
                "medium-large decoder, adapted as a prefix language model."
            ),
            "params": 2200345344,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_ml_ml_prefixlm/1",
    },
    "t5gemma_ml_ml_ul2_it": {
        "metadata": {
            "description": (
                "T5Gemma ML/ML model with a medium-large encoder and "
                "medium-large decoder, adapted as a UL2 model and fine-tuned "
                "for instruction following."
            ),
            "params": 2200345344,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_ml_ml_ul2_it/1",
    },
    "t5gemma_ml_ml_prefixlm_it": {
        "metadata": {
            "description": (
                "T5Gemma ML/ML model with a medium-large encoder and "
                "medium-large decoder, adapted as a prefix language model and "
                "fine-tuned for instruction following."
            ),
            "params": 2200345344,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_ml_ml_prefixlm_it/1",
    },
    "t5gemma_xl_xl_ul2": {
        "metadata": {
            "description": (
                "T5Gemma XL/XL model with an extra-large encoder and "
                "extra-large decoder, adapted as a UL2 model."
            ),
            "params": 3766980608,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_xl_xl_ul2/1",
    },
    "t5gemma_xl_xl_prefixlm": {
        "metadata": {
            "description": (
                "T5Gemma XL/XL model with an extra-large encoder and "
                "extra-large decoder, adapted as a prefix language model."
            ),
            "params": 3766980608,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_xl_xl_prefixlm/1",
    },
    "t5gemma_xl_xl_ul2_it": {
        "metadata": {
            "description": (
                "T5Gemma XL/XL model with an extra-large encoder and "
                "extra-large decoder, adapted as a UL2 model and fine-tuned "
                "for instruction following."
            ),
            "params": 3766980608,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_xl_xl_ul2_it/1",
    },
    "t5gemma_xl_xl_prefixlm_it": {
        "metadata": {
            "description": (
                "T5Gemma XL/XL model with an extra-large encoder and "
                "extra-large decoder, adapted as a prefix language model and "
                "fine-tuned for instruction following."
            ),
            "params": 3766980608,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_xl_xl_prefixlm_it/1",
    },
    "t5gemma_2b_2b_ul2": {
        "metadata": {
            "description": (
                "T5Gemma 2B/2B model with a 2-billion-parameter encoder and "
                "2-billion-parameter decoder, adapted as a UL2 model."
            ),
            "params": 5596853760,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_2b_2b_ul2/1",
    },
    "t5gemma_2b_2b_prefixlm": {
        "metadata": {
            "description": (
                "T5Gemma 2B/2B model with a 2-billion-parameter encoder and "
                "2-billion-parameter decoder, adapted as a prefix language "
                "model."
            ),
            "params": 5596853760,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_2b_2b_prefixlm/1",
    },
    "t5gemma_2b_2b_ul2_it": {
        "metadata": {
            "description": (
                "T5Gemma 2B/2B model with a 2-billion-parameter encoder and "
                "2-billion-parameter decoder, adapted as a UL2 model and "
                "fine-tuned for instruction following."
            ),
            "params": 5596853760,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_2b_2b_ul2_it/1",
    },
    "t5gemma_2b_2b_prefixlm_it": {
        "metadata": {
            "description": (
                "T5Gemma 2B/2B model with a 2-billion-parameter encoder and "
                "2-billion-parameter decoder, adapted as a prefix language "
                "model and fine-tuned for instruction following."
            ),
            "params": 5596853760,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_2b_2b_prefixlm_it/1",
    },
    "t5gemma_9b_2b_ul2": {
        "metadata": {
            "description": (
                "T5Gemma 9B/2B model with a 9-billion-parameter encoder and "
                "2-billion-parameter decoder, adapted as a UL2 model."
            ),
            "params": 12292375296,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_9b_2b_ul2/1",
    },
    "t5gemma_9b_2b_prefixlm": {
        "metadata": {
            "description": (
                "T5Gemma 9B/2B model with a 9-billion-parameter encoder and "
                "2-billion-parameter decoder, adapted as a prefix language "
                "model."
            ),
            "params": 12292375296,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_9b_2b_prefixlm/1",
    },
    "t5gemma_9b_2b_ul2_it": {
        "metadata": {
            "description": (
                "T5Gemma 9B/2B model with a 9-billion-parameter encoder and "
                "2-billion-parameter decoder, adapted as a UL2 model and "
                "fine-tuned for instruction following."
            ),
            "params": 12292375296,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_9b_2b_ul2_it/1",
    },
    "t5gemma_9b_2b_prefixlm_it": {
        "metadata": {
            "description": (
                "T5Gemma 9B/2B model with a 9-billion-parameter encoder and "
                "2-billion-parameter decoder, adapted as a prefix language "
                "model and fine-tuned for instruction following."
            ),
            "params": 12292375296,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_9b_2b_prefixlm_it/1",
    },
    "t5gemma_9b_9b_ul2": {
        "metadata": {
            "description": (
                "T5Gemma 9B/9B model with a 9-billion-parameter encoder and "
                "9-billion-parameter decoder, adapted as a UL2 model."
            ),
            "params": 20333401088,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_9b_9b_ul2/1",
    },
    "t5gemma_9b_9b_prefixlm": {
        "metadata": {
            "description": (
                "T5Gemma 9B/9B model with a 9-billion-parameter encoder and "
                "9-billion-parameter decoder, adapted as a prefix language "
                "model."
            ),
            "params": 20333401088,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_9b_9b_prefixlm/1",
    },
    "t5gemma_9b_9b_ul2_it": {
        "metadata": {
            "description": (
                "T5Gemma 9B/9B model with a 9-billion-parameter encoder and "
                "9-billion-parameter decoder, adapted as a UL2 model and "
                "fine-tuned for instruction following."
            ),
            "params": 20333401088,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_9b_9b_ul2_it/1",
    },
    "t5gemma_9b_9b_prefixlm_it": {
        "metadata": {
            "description": (
                "T5Gemma 9B/9B model with a 9-billion-parameter encoder and "
                "9-billion-parameter decoder, adapted as a prefix language "
                "model and fine-tuned for instruction following."
            ),
            "params": 20333401088,
            "path": "t5gemma",
        },
        "kaggle_handle": "kaggle://keras/t5-gemma/keras/t5gemma_9b_9b_prefixlm_it/1",
    },
}
