# Metadata for loading pretrained model weights.
backbone_presets = {
    "hgnetv2_b4_ssld_stage2_ft_in1k": {
        "metadata": {
            "description": (
                "HGNetV2 B4 model with 2-stage SSLD training, fine-tuned on "
                "ImageNet-1K."
            ),
            "params": 13599072,
            "path": "hgnetv2",
        },
        "kaggle_handle": "kaggle://keras/hgnetv2/keras/hgnetv2_b4_ssld_stage2_ft_in1k/1",
    },
    "hgnetv2_b5_ssld_stage1_in22k_in1k": {
        "metadata": {
            "description": (
                "HGNetV2 B5 model with 1-stage SSLD training, pre-trained on "
                "ImageNet-22K and fine-tuned on ImageNet-1K."
            ),
            "params": 33419680,
            "path": "hgnetv2",
        },
        "kaggle_handle": "kaggle://keras/hgnetv2/keras/hgnetv2_b5_ssld_stage1_in22k_in1k/1",
    },
    "hgnetv2_b5_ssld_stage2_ft_in1k": {
        "metadata": {
            "description": (
                "HGNetV2 B5 model with 2-stage SSLD training, fine-tuned on "
                "ImageNet-1K."
            ),
            "params": 33419680,
            "path": "hgnetv2",
        },
        "kaggle_handle": "kaggle://keras/hgnetv2/keras/hgnetv2_b5_ssld_stage2_ft_in1k/1",
    },
    "hgnetv2_b6_ssld_stage1_in22k_in1k": {
        "metadata": {
            "description": (
                "HGNetV2 B6 model with 1-stage SSLD training, pre-trained on "
                "ImageNet-22K and fine-tuned on ImageNet-1K."
            ),
            "params": 69179888,
            "path": "hgnetv2",
        },
        "kaggle_handle": "kaggle://keras/hgnetv2/keras/hgnetv2_b6_ssld_stage1_in22k_in1k/1",
    },
    "hgnetv2_b6_ssld_stage2_ft_in1k": {
        "metadata": {
            "description": (
                "HGNetV2 B6 model with 2-stage SSLD training, fine-tuned on "
                "ImageNet-1K."
            ),
            "params": 69179888,
            "path": "hgnetv2",
        },
        "kaggle_handle": "kaggle://keras/hgnetv2/keras/hgnetv2_b6_ssld_stage2_ft_in1k/1",
    },
}
